/*----------------------------------
         Midi68k v0.5
     Developed by Malcolm Smith
           Open Source
----------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <process.h>
#include <dos.h>
#include <time.h>
#define MAX_NOTES 16000 //Change this as needed.
//The limit for -O0 is 16011 notes. For -Os, it is 16066 notes.





int main(int argc, char *argv[])
{
short debug=0;
short pause=1;

FILE *in,*out;
const char * const logname="log.txt";
FILE *log;
short a;
short ties;
short unit=0;
short tempvar,tempvar2;
short tempo=0;
short restcount=0;
short chordcount=0;
double freqtable[128];
unsigned char buffer[80];
unsigned char buffer2[80];
unsigned int count=0;
unsigned int allmidi[MAX_NOTES][2]={0};
const char *dataname = "data.txt";
//Child function for running MIDI2TXT
void spawnl_midi2txt(void)
{
   int result;

   result = spawnl(P_WAIT, "midi2txt.exe", "midi2txt.exe","-dec",argv[1], dataname,NULL);
   if (result == -1)
   {
      perror("\nError from spawnl (Error using MIDI2TXT).");
//Display Usage Information
printf("

Midi68k Tool By Malcolm Smith
-----------------------------
Assembles TXT Midi Data into a TIGCC .c file for compilation.

USAGE:
midi68k midifile outfile \"title\" [-log -nopause]

midifile = Input MIDI file for compilation
        Example:song.mid
outfile = Desired name to write to, normally a C file.
        Example: music.c
title = Song Title/Author/Info. Must be in quotes, and must be <=26 characters.
        Example: \"Bob Smith - Techno Mix 1\"
-log = With this switch defined, a detailed log is kept in log.txt.
-nopause = With this switch defined, Midi68k does not pause before exiting.\n");
      exit(1);
   }
}
if(argc>=6)
{
if(!strcmp(argv[5],"-nopause"))  pause=0;
if(!strcmp(argv[5],"-log"))  debug=1;
}
if(argc>=5)
{
if(!strcmp(argv[4],"-log"))  debug=1;
if(!strcmp(argv[4],"-nopause"))  pause=0;
}
if(debug) {
printf("\nA detailed log will be kept in log.txt.\nFor normal operation, do not use the \"-log\" switch.\n\n");
}

//If not enough args (midifile,outfile,title) are given...
if(argc!=4 && !(argc==5 && debug) && !(argc==5 && !pause) && !(argc==6 && !pause))
{
printf("Incorrect number of args.\n");
//Display Usage Information
goto error;
}
//If the title string is too long...
if(strlen(argv[3])>26)
{
 printf("Title string is limited to 26 characters!\n");
 goto error;
}
//Run the MIDI2TXT program, opening midifile and outputting to data.txt
spawnl_midi2txt();
if(debug) {
log = fopen(logname,"ab");
//If not successful...
if(!log)
{
printf("Error opening %s for logging.\n", logname);
 goto error;
}
{
   struct tm *time_now;
   time_t secs_now;
   char str[80];

   tzset();
   time(&secs_now);
   time_now = localtime(&secs_now);

strftime (buffer, 80, "%x %I:%M:%S:%p",time_now);


fprintf(log,"\r\n\r\n---Begin Midi68k Log for %s---\r\n",argv[1]);
fprintf(log,"---Midi68k run at: %s---\r\n\r\n",buffer);

}
}
//Open data.txt for editing
in = fopen(dataname,"rb");
//If not successful...
if(!in)
{
printf("Error opening %s\n", dataname);
if(debug) {
fprintf(log,"Error opening %s\r\n", dataname);
}
 goto error;
}
//Read in the first four lines.
//In a true Type-0 MIDI file this is just header information.
for(a=0;a<4;a++)
{
	if (fgets(buffer, 80, in) == NULL) {
		printf("Error reading opening data (1st 4 lines).\n");
  if(debug) {
		fprintf(log,"Error reading opening data (1st 4 lines).\r\n");
        }
		fclose(in);
		goto error;
	}
}
// The 5th line is the 'unit' line.
// This gives a divisor for all of the other durations.
// If it can't be read, display usage info and quit.
	if (fgets(buffer, 80, in) == NULL) {
		printf("Error reading UNIT line.\n");
if(debug) {
fprintf(log,"Error reading UNIT line.\r\n");
}
		fclose(in);
		goto error;
	}
//Save this unit information
 sscanf(buffer, "  unit %i", &tempvar);
 unit=tempvar;
// This keeps reading lines in until it finds the
// 'beats' (tempo) line. If it runs out of file it
// returns an error.
while(buffer[2]!='b')//i.e. "beats"
{
	if (fgets(buffer, 80, in) == NULL) {
		printf("Error reading pre-MIDI data.\n");
if(debug) {
fprintf(log,"Error reading pre-MIDI data.\r\n");
}
		fclose(in);
		goto error;
	}
}
// Saves the beat/tempo information
 sscanf(buffer, "  beats %i", &tempvar);
 tempo=tempvar;
// This keeps reading lines in until it finds the
// first MIDI data line. If it runs out of file it
// returns an error.
while(buffer[2]!='+' && buffer[5]!='+' &&buffer[6]!='+' &&buffer[7]!='+')//i.e. the MIDI data
{
	if (fgets(buffer, 80, in) == NULL) {
		printf("Error reading pre-MIDI data.\n");
if(debug) {
fprintf(log,"Error reading pre-MIDI data.\r\n");
}
		fclose(in);
		goto error;
	}
// If there's another tempo line, change the tempo.
// Some software always begins with tempo at 120,
// and afterwards changes it to what the user
// wanted, but before the MIDI data starts.
 if(buffer[2]=='b') //i.e. beats, a second time
  {  sscanf(buffer, "  beats %i", &tempvar);
 tempo=tempvar;
 }

}
// Displays the tempo for the user
printf("\nTempo is %i.",tempo);
// Displays the unit for the user
printf("\nUnit is %i.",unit);
if(debug) {
fprintf(log,"\r\nTempo is %i.",tempo);
fprintf(log,"\r\nUnit is %i.",unit);
}
// Generate the frequency table
// This table is used to look up MIDI notes
// and find their corresponding frequencies.
for(a=1;a<129;a++)
{
short newa=a-9;
double val=(double)(newa)/12;
double val2 = (double)pow((double)2,val);
double val3 = (double)val2*(double)(440/32);
freqtable[a-1] =  val3;
}
//The main MIDI-reading loop...
while(1)
{
chordcount=0;
restcount=0;
ties=0;
if(count!=0) // If it's the first time, a line has already been read!
{
 // If we ran out of data...
  if (fgets(buffer, 80, in) == NULL) break;
}
  // If there are too many notes to read...
  if(count==MAX_NOTES) {
  printf("\nToo many notes, bailing out.\nTruncating file to avoid the 64k limit.\nThe limit on notes is %i.",MAX_NOTES);
  if(debug) {
  fprintf(log,"\r\nToo many notes, bailing out.\r\nTruncating file to avoid the 64k limit.\r\nThe limit on notes is %i.",MAX_NOTES);
  }
  break;}
loopbeg:
  //OK, no problems, let's continue.
  //Check the first line for tempo changes, and time signature changes.
  {
     short check=2;
     short flag=0;
     if(buffer[2]>='0' && buffer[2]<='9')
     //Hm, there is a rest.
     {
       tempvar=0;
         sscanf(buffer, "  %i", &tempvar);
           //Convert the duration to a usable form
           {
           double duration = (double)tempvar/unit;
           double val = (double)60/tempo;//tempos are in Beats/Minute
           double val2 = (double)val*duration*350;
           double final = ceil(val2); //ceil() always rounds up (so no durations will be 0).
            //Put the duration in the array
             allmidi[count][0] = (short)final;
             allmidi[count][1] = 0;
             count++;
             restcount++;
           }
       while(buffer[check]!=';') check++;
       check++;
     }
     if((!strncmp(buffer+check,"tact",4)) || (!strncmp(buffer+check,"key",3)))
     //A time signature change/ key change
      flag++;  //Ignore it, it doesn't matter.
     else if(!strncmp(buffer+check,"beats",5))  //A tempo change
     {
     tempvar=0;
     sscanf(buffer+check+6, "%i", &tempvar);
     tempo=tempvar;
     flag++;
     }
     if(flag) {fgets(buffer, 80, in);goto loopbeg;}
  }
  //The first line doesn't really matter unless the note before was a rest.
  //Read the second line, too.
    fgets(buffer2, 80, in);
  if(buffer[2]>='0' && buffer[2]<='9' && !restcount)
  //If there is a delay first (a rest) and it isn't a chord, and it wasn't already found
  {
  tempvar=0;
  sscanf(buffer, "  %i", &tempvar);
  //Convert the duration to a usable form
{
double duration = (double)tempvar/unit;
double val = (double)60/tempo;
double val2 = (double)val*duration*350;
double final = ceil(val2); //ceil() always rounds up (so no durations will be 0).
  //Put the duration in the array
  allmidi[count][0] = (short)final;
}
  allmidi[count][1]=0;
  count++;
  restcount++;
  }
  //If the MIDI file contains a chord, the first line DOES matter!
  //If the second line begins with a plus, it's a chord.
  if(buffer2[2]=='+')
  {
  short a;
  short tempvar3;
  //Chord handling. Note that all notes besides the first are eliminated.
  //Now, go back and look at the first line.
  //Scan in the duration and MIDI note
  tempvar=0;
  tempvar2=0;
  sscanf(buffer, "  +%i", &tempvar2);
 //  }
  chordcount=2;
  //Ignore the other notes in the chord
  while(buffer2[2]=='+'||buffer2[6]=='+')
  {
  fgets(buffer2, 80, in);
  chordcount++;
  }
  sscanf(buffer2, "  %i;-%i 0;", &tempvar,&tempvar2);
  sscanf(buffer2, "  %i", &tempvar);
  for(a=1;a<chordcount-1;a++)
  {
   fgets(buffer2, 80, in);
   if(buffer2[2]!='-' && buffer2[6]!='-')//Hm, another note is going on! Uh oh.
   {ties=1;goto loopbeg;}
  }
  }
  else //If it isn't a chord...
  {
  //Scan in the duration and MIDI note
  tempvar=0;
  tempvar2=0;
  sscanf(buffer2, "  %i;-%i 0;", &tempvar,&tempvar2);
  sscanf(buffer2, "  %i", &tempvar);
  }
  //Convert the MIDI note to a real frequency for on-calc playing
  allmidi[count][1] = (short)freqtable[tempvar2];
//Convert the duration to a usable form
{
double duration = (double)tempvar/unit;
double val = (double)5400/tempo;//(5400 is 27x200, i.e. a quarter note is 27 long at tempo 200)
double val2 = (double)val*duration;
double final = ceil(val2); //ceil() always rounds up (so no durations will be 0).
  //Put the duration in the array
  allmidi[count][0] = (short)final;
if(debug) {
//Log the input
  fprintf(log,"\r\n Input: %s",buffer2);
//Log the output
  fprintf(log,"Output: %i %i",tempvar,tempvar2);
  fprintf(log,"\r\nAllMidi Data:%i %i\r\n",allmidi[count][0],allmidi[count][1]);
}
}
  //Continue
  count++;

    if(ties)
  {
   ties=0;
   fgets(buffer2, 80, in);
   if(buffer2[2]!='-' && buffer2[5]!='-' && buffer2[6]!='-' && buffer2[7]!='-' && buffer2[8]!='-')
   //Hm, another note is going on! Uh oh.
   {if(!strncmp(buffer2,"end mtrk",8)) break;
   ties=1;goto loopbeg;}
  }

}
//The loop is over, and count is one too high.
//Fix the count variable.
count--;

if(debug) {
fprintf(log,"\r\n\r\nBeginning log of allmidi array.\r\n");
for(a=0;a<count;a++)
{
fprintf(log,"\r\n%i %i",allmidi[a][0],allmidi[a][1]);
}
}
// Close data.txt
fclose(in);
// Open the 'outfile' to write the array and code to.
out = fopen(argv[2],"wt");
//Write the new data to a file.
fprintf(out,
"#define SAVE_SCREEN
#define NO_EXIT_SUPPORT
#define USE_TI89
#define USE_TI92PLUS
#define MIN_AMS 200           // Compile for AMS 2.00 or higher
#include <tigcclib.h>         // Include all header files

");
fprintf(out,"short int music[%i][2]={\n",count+2);
for(a=0;a<(count-1);a++)
{
short cur=allmidi[a][0];
if(cur>0) fprintf(out,"{%i,%i},\n",allmidi[a][0],allmidi[a][1]);
else {fprintf(out,"{0,0},\n");}
}
a++;
if(allmidi[count-1][0]>0) fprintf(out,"{%i,%i}",allmidi[count-1][0],allmidi[count-1][1]);
else fprintf(out,"{0,0}");
fprintf(out,"};\n");
fprintf(out,"

// This contains the HW2 status.
// 0 for HW1s, 1 for HW2s. This is used because they are
// different for sound.
short hardwarever;

// This works through Auto-Int-5, as an always-going-counter.
volatile short count;
DEFINE_INT_HANDLER(count_int)
{
	++count;
}


//=============================================================================
// VTI detection routine (thanx, Julien Muchembled)
//
// this routine utilizes a bug in VTI: you can set even non-existing flags in
// the status register and query them later!
//=============================================================================
extern short IsVTI(void);
asm(\"IsVTI:    trap   %c12;         /* enter supervisor mode. returns old (%%sr) in %%d0.w   */    move.w %c0x3000,%%sr; /* set a non-existing flag in %%sr (but keep s-flag !!) */    swap   %%d0;         /* save %%d0.w content in upper part of %%d0             */    move.w %%sr,%%d0;     /* get %%sr content and check for non-existing flag     */    btst   %c12,%%d0;     /* this non-existing flag can only be set on the VTI   */    bne    __VTI;    swap   %%d0;         /* restore old %%sr content and return 0               */    move.w %%d0,%%sr;    moveq  %c0,%%d0;    rts;__VTI:    swap   %%d0;         /* restore old %%sr content and return 1                */    move.w %%d0,%%sr;     moveq  %c1,%%d0;     rts;\");
short VTI;


void rest(long duree)
{
	short start = count;
	while (count - start < duree);
}
//Thanks to Francois Leiber for the beep() code.
//However, it has been slightly modified with ASM
//so it works with VTI's sound feature.
void beep(short freq, long duree)
{

	long int loop, maxloop ;
 short int delay;
	short int i, j;
if(VTI) duree=(duree*15)>>4;
maxloop = freq*duree/50;
j = peekIO(0x60000E);
freq = (hardwarever ? 364000 : 278000)/freq ;
asm(\"bset.b		#6,0x60000C\");
	for (loop = 0 ; loop < maxloop ; loop++)
	{
asm(\"bchg.b		#0,0x60000E;bchg.b		#1,0x60000E;\");
		for (i = 0 ; i <= freq ; i--) i += 2 ;
	}
asm(\"bclr.b 		#0,0x60000E;bclr.b 		#1,0x60000E;move.b 		#0x8D,0x60000C;\");
for(delay=0;delay<(100*duree);delay+=2) delay--;
}
",35,35,35,35,35);   //35 is the '#' symbol
fprintf(out,"

void _main(void)
{
  INT_HANDLER save_int_4 = GetIntVec(AUTO_INT_4);
  INT_HANDLER save_int_1 = GetIntVec(AUTO_INT_1);
  INT_HANDLER save_int_5 = GetIntVec(AUTO_INT_5);
  short int total = sizeof(music)/sizeof(music[0]);
  short int current=0;
  /* Checks the calculator hardware version */
	short *hwptr;
	hwptr = (short *)FL_getHardwareParmBlock();
 hardwarever = ((*hwptr)==20)?0:1;
 /*-----------------------------------------*/
 VTI=IsVTI();
 //Directs ints to nothing
  SetIntVec(AUTO_INT_4, DUMMY_HANDLER);
  SetIntVec(AUTO_INT_5, DUMMY_HANDLER);
  //Auto Int 1 is called 350 times per second.
  SetIntVec(AUTO_INT_1, count_int);
  OSLinkReset();
  OSLinkOpen();
  clrscr();
printf(\"\\n\\n%s\");",argv[3]);
fprintf(out,"
printf(\"\\n\\n\\nCreated by Midi68k\\nBy Malcolm Smith\\nhttp://jesystems.cjb.net\\n\\nHOLD Esc to Quit.\");
  while( !(_keytest (RR_ESC)) && current<total)
  {
  if(!music[current][1]) rest(music[current][0]);
  else if(music[current][0]>0)  beep(music[current][1],music[current][0]);
  current++;
  }
	SetIntVec(AUTO_INT_4, save_int_4);
	SetIntVec(AUTO_INT_1, save_int_1);
	SetIntVec(AUTO_INT_5, save_int_5);
 OSLinkReset();
 GKeyFlush();
 OSTimerRestart(APD_TIMER);
}

");
// OK, now the code and array are written.
// Close the file.
fclose(out);
// Delete the "intermediate" data.txt.
remove(dataname);
// Display the success message.
printf("\n\nMidi68k successfully created %s.\n\nCompile with TIGCC (http:\x2f\x2ftigcc.ticalc.org) to make a .89z file.\n",argv[2]);
if(debug) {
fprintf(log,"\r\n\r\nMidi68k successfully created %s.\r\n\r\nCompile with TIGCC (http:\x2f\x2ftigcc.ticalc.org) to make a .89z file.\r\n",argv[2]);
fclose(log);
}
//Quit. Bye.
if(pause) system("PAUSE");
return 0;

// This is where direction is sent when an error occurs.
// This displays usage info and returns a nonzero.
error:
//Display Usage Information
printf("
Error!

Midi68k Tool By Malcolm Smith
-----------------------------
Assembles TXT Midi Data into a TIGCC .c file for compilation.

USAGE:
midi68k midifile outfile \"title\" [-log -nopause]

midifile = Input MIDI file for compilation
        Example:song.mid
outfile = Desired name to write to, normally a C file.
        Example: music.c
title = Song Title/Author/Info. Must be in quotes, and must be <=26 characters.
        Example: \"Bob Smith - Techno Mix 1\"
-log = With this switch defined, a detailed log is kept in log.txt.
-nopause = With this switch defined, Midi68k will not pause before exiting.

");
remove(dataname);
if(debug) {
if(log) fclose(log);
}
return 1;
}
